from torch.utils.data import DataLoader as TrochDataLoader


class Dataset():
    def __init__(self, x, y):
        self.x, self.y = x, y

    def __len__(self):
        return len(self.x)

    def __getitem__(self, i):
        return self.x[i], self.y[i]


class DataLoader(TrochDataLoader):
    def __init__(self, *p, **pd):
        super().__init__(*p, **pd)

    @classmethod
    def loaders(cls, train_ds, val_ds, test_ds=None, batch_size=1):
        train_dl = cls(train_ds, batch_size=batch_size)
        val_dl = cls(val_ds, batch_size=batch_size)
        if test_ds is None:
            return train_dl, val_dl
        else:
            return train_dl, val_dl, cls(test_ds, batch_size=batch_size)


class DataBunch():
    def __init__(self, train_dl, valid_dl=None):
        self.train_dl, self.valid_dl = train_dl, valid_dl

    @classmethod
    def from_dataset(cls, train_ds, valid_ds=None, batch_size=1):
        train_dl = DataLoader(train_ds, batch_size=batch_size)
        if valid_ds is not None:
            valid_dl = DataLoader(valid_dl, batch_size=batch_size)
        else:
            valid_dl = None
        return cls(train_dl, valid_dl)

    @property
    def train_ds(self):
        return self.train_dl.dataset

    @property
    def valid_ds(self):
        if self.valid_dl is None:
            return None
        return self.valid_dl.dataset


class DataProcess:
    """
    数据处理工具函数
    """
    @classmethod
    def normalize(cls, x, m, s):
        """
        标准化
        """
        return (x-m)/s

    @classmethod
    def stats(cls, x):
        """
        均值，标准差
        """
        return x.mean(), x.std()

    @classmethod
    def get_dls(cls, train_ds, valid_ds, bs, **kwargs):
        return (DataLoader(train_ds, batch_size=bs, shuffle=True, **kwargs),
                DataLoader(valid_ds, batch_size=bs*2, **kwargs))

    @classmethod
    def normalize_to(cls, train, valid, test=None):
        m, s = train.mean(), train.std()
        if test is None:
            return cls.normalize(train, m, s), cls.normalize(valid, m, s)
        else:
            return cls.normalize(train, m, s), cls.normalize(valid, m, s), cls.normalize(test, m, s)

    @classmethod
    def flatten(cls, x):
        return x.view(x.shape[0], -1)



import torch
from torch_geometric.data import DataLoader as DataLoader_pyg

# from .data import DataBunch, PyGDataLoaderWrapper


class PyGDataLoader(DataLoader_pyg):

    def __iter__(self):
        for data in super().__iter__():
            yield data, data.y

    @classmethod
    def loaders(cls, train_ds, val_ds, test_ds=None, batch_size=1):
        train_dl = cls(train_ds, batch_size=batch_size)
        val_dl = cls(val_ds, batch_size=batch_size)
        if test_ds is None:
            return train_dl, val_dl
        else:
            return train_dl, val_dl, cls(test_ds, batch_size=batch_size)


class MaskedTensor:
    def __init__(self, data, mask):
        self.data = data
        self.mask = mask

    def to(self, device):
        self.data = self.data.to(device)
        self.mask = self.mask.to(device)
        return self

    def __len__(self):
        return len(self.data)


class PyGGraphLoader:
    """
    用于节点计算，仅包含单一的torch_geometric.data.data.Data的Wrapper
    """

    def __init__(self, pyg_dataset, phase):
        """
        Args:
            pyg_data: torch_geometric.data.data.Data对象
            phase: 'train', 'val' or 'test'
        """
        pyg_data = pyg_dataset[0]
        assert phase in ['train', 'val', 'test'], 'param "phase" must be in [train, val, test]'
        assert pyg_data.x is not None, 'Data must has a attribute "x"'
        assert pyg_data.y is not None, 'Data must has a attribute "y"'
        self.mask = getattr(pyg_data, f'{phase}_mask')
        assert self.mask is not None, f'Data must has a attribute "{phase}_mask"'

        self.data = pyg_data
        self.phase = phase

    @classmethod
    def loaders(cls, pyg_dataset, phases=('train', 'val', 'test')):
        return [cls(pyg_dataset, phase) for phase in phases]

    def __iter__(self):
        yield self.data, MaskedTensor(self.data.y, self.mask)

    def __len__(self):
        if torch.sum(self.mask.int()) > 0:
            return 1
        else:
            return 0
